// Copyright (C) 2018-2021 Internet Systems Consortium, Inc. ("ISC")
//
// This Source Code Form is subject to the terms of the End User License
// Agreement. See COPYING file in the premium/ directory.

#include <config.h>

#include <legal_log_log.h>
#include <backend_store.h>

#include <rotating_file.h>
#ifdef HAVE_MYSQL
#include <mysql_legal_log.h>
#endif
#ifdef HAVE_PGSQL
#include <pgsql_legal_log.h>
#endif
#ifdef HAVE_CQL
#include <cql_legal_log.h>
#endif

#include <database/database_connection.h>
#include <dhcpsrv/cfgmgr.h>
#include <eval/eval_context.h>

#include <boost/date_time/posix_time/posix_time.hpp>

#include <errno.h>
#include <iostream>
#include <sstream>
#include <time.h>

namespace isc {
namespace legal_log {

using namespace isc::asiolink;
using namespace isc::data;
using namespace isc::db;
using namespace isc::dhcp;
using namespace isc::eval;
using namespace isc::hooks;


void
BackendStore::parseFile(const ConstElementPtr& parameters) {
    // Defaults
    std::string path(LEGAL_LOG_DIR);
    std::string base("kea-legal");
    RotatingFile::TimeUnit unit(RotatingFile::TimeUnit::Day);
    int64_t count(1);
    std::string prerotate;
    std::string postrotate;

    // Prioritize parameters.
    if (parameters) {
        if (parameters->get("path")) {
            path = parameters->get("path")->stringValue();
        }
        if (parameters->get("base-name")) {
            base = parameters->get("base-name")->stringValue();
        }
        if (parameters->get("time-unit")) {
            std::string time_unit(parameters->get("time-unit")->stringValue());
            if (time_unit == "second") {
                unit = RotatingFile::TimeUnit::Second;
            } else if (time_unit == "day") {
                unit = RotatingFile::TimeUnit::Day;
            } else if (time_unit == "month") {
                unit = RotatingFile::TimeUnit::Month;
            } else if (time_unit == "year") {
                unit = RotatingFile::TimeUnit::Year;
            } else {
                isc_throw(BadValue, "unknown time unit type: " << time_unit
                            << ", expected one of: second, day, month, year");
            }
        }
        if (parameters->get("count")) {
            count = parameters->get("count")->intValue();
            if ((count < 0) ||
                (count > std::numeric_limits<uint32_t>::max())) {
                isc_throw(OutOfRange, "count value: " << count
                        << " is out of range, expected value: 0.."
                        << std::numeric_limits<uint32_t>::max());
            }
        }
        if (parameters->get("prerotate")) {
            prerotate = parameters->get("prerotate")->stringValue();
        }
        if (parameters->get("postrotate")) {
            postrotate = parameters->get("postrotate")->stringValue();
        }
    }

    BackendStore::instance() = boost::make_shared<RotatingFile>(
        path, base, unit, count, prerotate, postrotate);
}

void
BackendStore::parseDatabase(const ConstElementPtr& parameters) {
    // Should never happen with the code flow at the time of writing, but
    // let's get this check out of the way.
    if (!parameters) {
        isc_throw(BadValue, "no parameters specified for the hook library");
    }

    DatabaseConnection::ParameterMap db_parameters;

    // Strings
    for (char const* const& key : {
         "type", "user", "password", "host", "name", "contact-points",
         "keyspace", "consistency", "serial-consistency"}) {
        ConstElementPtr const value(parameters->get(key));
        if (value) {
            db_parameters.emplace(key, value->stringValue());
        }
    }

    // uint32_t
    for (char const* const& key : {
         "connect-timeout", "reconnect-wait-time", "max-reconnect-tries",
         "request-timeout", "tcp-keepalive"}) {
        ConstElementPtr const value(parameters->get(key));
        if (value) {
            int64_t integer_value(value->intValue());
            auto const max(std::numeric_limits<uint32_t>::max());
            if (integer_value < 0 || max < integer_value) {
                isc_throw(OutOfRange,
                          key << " value: " << integer_value
                              << " is out of range, expected value: 0.."
                              << max);
            }
            db_parameters[key] =
                boost::lexical_cast<std::string>(integer_value);
        }
    }

    std::string param_name = "tcp-nodelay";
    ConstElementPtr param = parameters->get(param_name);
    if (param) {
        db_parameters.emplace(param_name,
                              param->boolValue() ? "true" : "false");
    }

    // Always set "on-fail" to "serve-retry-continue" if not explicitly
    // configured.
    std::string on_fail_action = ReconnectCtl::onFailActionToText(OnFailAction::SERVE_RETRY_CONTINUE);
    param_name = "on-fail";
    param = parameters->get(param_name);
    if (param) {
        on_fail_action = param->stringValue();
        ReconnectCtl::onFailActionFromText(on_fail_action);
    }
    db_parameters.emplace(param_name, on_fail_action);

    int64_t port = 0;
    param_name = "port";
    param = parameters->get(param_name);
    if (param) {
        port = param->intValue();
        if ((port < 0) || (port > std::numeric_limits<uint16_t>::max())) {
            isc_throw(OutOfRange, param_name << " value: " << port
                      << " is out of range, expected value: 0.."
                      << std::numeric_limits<uint16_t>::max());
        }
        db_parameters.emplace(param_name,
                              boost::lexical_cast<std::string>(port));
    }

    std::string redacted =
        DatabaseConnection::redactedAccessString(db_parameters);

    std::string const dbtype(db_parameters["type"]);
#ifdef HAVE_MYSQL
    if (dbtype == "mysql") {
        LOG_INFO(legal_log_logger, LEGAL_LOG_MYSQL_DB).arg(redacted);
        BackendStore::instance().reset(new MySqlStore(db_parameters));
        return;
    }
#endif
#ifdef HAVE_PGSQL
    if (dbtype == "postgresql") {
        LOG_INFO(legal_log_logger, LEGAL_LOG_PGSQL_DB).arg(redacted);
        BackendStore::instance().reset(new PgSqlStore(db_parameters));
        return;
    }
#endif
#ifdef HAVE_CQL
    if (dbtype == "cql") {
        LOG_INFO(legal_log_logger, LEGAL_LOG_CQL_DB).arg(redacted);
        BackendStore::instance().reset(new CqlStore(db_parameters));
        return;
    }
#endif

    isc_throw(InvalidType, "Database access parameter 'type' does "
              "not specify a supported database backend: " << dbtype);
}

void
BackendStore::parseExtraParameters(const ConstElementPtr& parameters) {
    if (!parameters) {
        return;
    }
    ConstElementPtr param = parameters->get("request-parser-format");
    if (param && !param->stringValue().empty()) {
        BackendStore::instance()->setRequestFormatExpression(param->stringValue());
    }

    param = parameters->get("response-parser-format");
    if (param && !param->stringValue().empty()) {
        BackendStore::instance()->setResponseFormatExpression(param->stringValue());
    }
}

IOServicePtr BackendStore::io_service_;
DatabaseConnection::ParameterMap BackendStore::parameters_;

struct tm
BackendStore::currentTimeInfo() const {
    struct tm time_info;
    time_t timestamp = now();
    localtime_r(&timestamp, &time_info);
    return (time_info);
}

time_t
BackendStore::now() const {
    return (time(NULL));
}

std::string
BackendStore::getNowString(const std::string& format) const {
    // Get a text representation of the current time.
    return (getTimeString(now(), format));
}

std::string
BackendStore::getTimeString(time_t time, const std::string& format) {
    // Get a text representation of the requested time.
    char buffer[128];
    struct tm time_info;
    localtime_r(&time, &time_info);

    if (!strftime(buffer, sizeof(buffer), format.c_str(), &time_info)) {
        isc_throw(BackendStoreError,
                  "Timestamp format format: " << format
                  << " result is too long, maximum allowed: "
                  << sizeof(buffer));
    }
    return (std::string(buffer));
}

std::string
BackendStore::genDurationString(const uint32_t secs) {
    // Because Kea handles lease lifetimes as uint32_t and supports
    // a value of 0xFFFFFFFF (infinite lifetime), we don't use things like
    // boost:posix_time::time_duration as they work on longs.  Therefore
    // we'll figure it out ourselves.  Besides, the math ain't that hard.
    if (secs == 0xffffffff) {
        return ("infinite duration");
    }

    uint32_t seconds = secs % 60;
    uint32_t remainder = secs / 60;
    uint32_t minutes = remainder % 60;
    remainder /= 60;
    uint32_t hours = remainder % 24;
    uint32_t days = remainder / 24;

    std::ostringstream os;
    // Only spit out days if we have em.
    if (days) {
        os << days << " days ";
    }

    os << hours << " hrs "
       << minutes << " mins "
       << seconds << " secs";

    return (os.str());
}

std::string
BackendStore::vectorHexDump(const std::vector<uint8_t>& bytes,
                         const std::string& delimiter) {
    std::stringstream tmp;
    tmp << std::hex;
    bool delim = false;
    for (std::vector<uint8_t>::const_iterator it = bytes.begin();
         it != bytes.end(); ++it) {
        if (delim) {
            tmp << delimiter;
        }
        tmp << std::setw(2) << std::setfill('0') << static_cast<unsigned int>(*it);
        delim = true;
    }
    return (tmp.str());
}

std::string
BackendStore::vectorDump(const std::vector<uint8_t>& bytes) {
    if (bytes.empty()) {
        return (std::string());
    }
    return (std::string(bytes.cbegin(), bytes.cend()));
}

void
BackendStore::setRequestFormatExpression(const std::string& extended_format) {
    Option::Universe universe;
    if (CfgMgr::instance().getFamily() == AF_INET) {
        universe = Option::V4;
    } else {
        universe = Option::V6;
    }
    EvalContext eval_ctx(universe);
    eval_ctx.parseString(extended_format, EvalContext::PARSER_STRING);
    request_expression_.reset(new Expression(eval_ctx.expression));
}

void
BackendStore::setResponseFormatExpression(const std::string& extended_format) {
    Option::Universe universe;
    if (CfgMgr::instance().getFamily() == AF_INET) {
        universe = Option::V4;
    } else {
        universe = Option::V6;
    }
    EvalContext eval_ctx(universe);
    eval_ctx.parseString(extended_format, EvalContext::PARSER_STRING);
    response_expression_.reset(new Expression(eval_ctx.expression));
}

const std::string
actionToVerb(Action action) {
    switch (action) {
    case Action::ASSIGN:
        return ("assigned");
    case Action::RELEASE:
        return ("released");
    default:
        return ("unknown-action");
    }
}

} // namespace legal_log
} // namespace isc
