// Copyright (C) 2017-2021 Internet Systems Consortium, Inc. ("ISC")
//
// This Source Code Form is subject to the terms of the End User License
// Agreement. See COPYING file in the premium/ directory.

/// @file command_callouts.cc Defines lease4_select and lease4_renew callout functions.

#include <config.h>
#include <cc/data.h>
#include <cc/simple_parser.h>
#include <dhcpsrv/cfgmgr.h>
#include <dhcpsrv/cfg_subnets4.h>
#include <dhcpsrv/cfg_subnets6.h>
#include <dhcpsrv/subnet.h>
#include <hooks/hooks.h>
#include <util/strutil.h>
#include <legal_log_log.h>
#include <backend_store.h>
#include <subnets_user_context.h>

#include <sstream>

using namespace isc;
using namespace data;
using namespace dhcp;
using namespace hooks;
using namespace util;
using namespace legal_log;
using namespace std;

// Functions accessed by the hooks framework use C linkage to avoid the name
// mangling that accompanies use of the C++ compiler as well as to avoid
// issues related to namespaces.
extern "C" {

/// @brief Fetch value for a string element if it is an element scope
///
/// Exception safe fetcher which looks for a string element with the
/// given name within an given element scope. Rather than throw if the
/// element is not found, it returns true if found, false if not.
///
/// @param arguments element scope to search, assumed to be a map
/// @param name name of the desired element
/// @param[out] value reference to a string in which the found value is
/// returned.  Will be set to "" if the element is not found.
///
/// @return true if the element was found, false if not
bool getOptionalString(ConstElementPtr& arguments, const std::string& name,
                       std::string& value) {
    value = "";
    try {
        value = SimpleParser::getString(arguments, name);
        return (true);
    } catch (...) {
        // we don't care why
    }

    return(false);
}

/// @brief Fetch value for a integer element if it is an element scope
///
/// Exception safe fetcher which looks for an integer element with the
/// given name within an given element scope. Rather than throw if the
/// element is not found, it returns true if found, false if not.
///
/// @param arguments element scope to search, assumed to be a map
/// @param name name of the desired element
/// @param[out] value reference to an integer in which the found value is
/// returned.  Will be set to 0 if the element is not found.
///
/// @return true if the element was found, false if not
bool getOptionalInt(ConstElementPtr& arguments, const std::string& name,
                    int64_t& value) {
    value = 0;
    try {
        value = SimpleParser::getInteger(arguments, name);
        return (true);
    } catch (...) {
        // we don't care why
    }

    return(false);
}

/// @brief Returns true if an element scope describes a prefix delegation
///
/// Searches the given scope of an lease type element, "type".  If it
/// exists and it's value is "IA_PD" or "2" return true, false otherwise.
///
/// @param arguments element scope to search, assumed to be a map
///
/// @return true if the scope describes a prefix delegation
bool isPrefix(ConstElementPtr arguments) {
    std::string type_str;
    if (getOptionalString(arguments, "type", type_str)) {
        return (type_str == "IA_PD" || type_str == "2");
    }

    return (false);
}

/// @brief Outputs text describing lease duration to a stream
///
/// Looks for either a lease valid lifetime or expiry from which to
/// generate the duration text using @ref
/// isc::legal_log::BackendStore::genDurationString().
///
/// @param os output stream to which the text is output
/// @param arguments element scope which may contain duration values
void addDuration(std::ostringstream& os, ConstElementPtr& arguments) {

    int64_t duration = 0;
    if (!getOptionalInt(arguments, "valid-lft", duration)) {
        int64_t expire = 0;
        if (getOptionalInt(arguments, "expire", expire)) {
            duration = expire - BackendStore::instance()->now();
        }
    }

    if (duration > 0) {
        os << " for " << BackendStore::genDurationString(duration);
    }
}

/// @brief Outputs text describing lease optional user context to a stream
///
/// Looks for an user context or a comment and display it
///
/// @param os output stream to which the text is output
/// @param arguments element scope which may contain user context or comment
void addContext(std::ostringstream& os, ConstElementPtr& arguments) {

    ConstElementPtr comment = arguments->get("comment");
    ConstElementPtr ctx = arguments->get("user-context");
    if (comment) {
        ElementPtr copied;
        if (ctx) {
            copied = copy(ctx, 0);
        } else {
            copied = Element::createMap();
        }
        copied->set("comment", comment);
        ctx = copied;
    }
    if (ctx) {
        os << ", context: " << ctx->str();
    }
}

/// @brief  This callout is called at the "command_processed" hook point.
///
/// Generates an entry in the legal log for the given control channel
/// command.
///
/// It is possible to disable logging for selected subnets by specifying a
/// "legal-logging" boolean parameter within the subnet's user context. If
/// this parameter is not specified it defaults to 'true', in which case the
/// logging is performed. When this value is set to 'false' the log is not
/// produced and the function returns.
///
/// @param handle CalloutHandle which provides access to context.
///
/// @return 0 upon success, non-zero otherwise.
int command_processed(CalloutHandle& handle) {
    if (!BackendStore::instance()) {
        LOG_ERROR(legal_log_logger,
                  LEGAL_LOG_COMMAND_NO_LEGAL_STORE);
        return (1);
    }

    std::string name;
    ConstElementPtr arguments;
    ConstElementPtr response;
    try {
        handle.getArgument("name", name);
        handle.getArgument("arguments", arguments);
        handle.getArgument("response", response);

        int result = SimpleParser::getInteger(response, "result");
        if (result != 0) {
            // We don't log failed commands
            return(0);
        }

        // We are only interested in the following commands.
        if ((name != "lease4-add") && (name != "lease4-update") && (name != "lease4-del") &&
            (name != "lease6-add") && (name != "lease6-update") && (name != "lease6-del")) {
            return (0);
        }

        // Check if the subnet identifier has been specified for the command. In some cases
        // it may be missing, i.e. lease4-del or lease6-del commands, when deleting a lease
        // by an IP address or prefix.
        int64_t subnet_id_value = 0;
        if (getOptionalInt(arguments, "subnet-id", subnet_id_value) && (subnet_id_value > 0)) {
            // The subnet identifier is present and valid.
            SubnetID subnet_id = static_cast<SubnetID>(subnet_id_value);

            // We need to get a different configuration pointer depending if we're in v4 or v6
            // universe. The easiest way to check the universe is by the command name.
            if (name.find("lease4-") != std::string::npos) {
                CfgSubnets4Ptr cfg = CfgMgr::instance().getCurrentCfg()->getCfgSubnets4();
                // Get the subnet by id and if the subnet is present, check if its user
                // context to see if legal logging is enabled for this subnet.
                ConstSubnet4Ptr subnet = cfg->getBySubnetId(subnet_id);
                if (isLoggingDisabled(subnet)) {
                    return (0);
                }

            } else if (name.find("lease6-") != std::string::npos) {
                CfgSubnets6Ptr cfg = CfgMgr::instance().getCurrentCfg()->getCfgSubnets6();
                // Get the subnet by id and if the subnet is present, check if its user
                // context to see if legal logging is enabled for this subnet.
                ConstSubnet6Ptr subnet = cfg->getBySubnetId(subnet_id);
                if (isLoggingDisabled(subnet)) {
                    return (0);
                }
            }
        }

        std::ostringstream osa;
        std::ostringstream os;
        os << "Administrator";

        if ((name == "lease4-add") || name == "lease4-update") {
            if (name == "lease4-add") {
                os << " added a lease of address: ";
            } else {
                os << " updated information on the lease of address: ";
            }
            osa << SimpleParser::getString(arguments, "ip-address");
            os << SimpleParser::getString(arguments, "ip-address")
               << " to a device with hardware address: "
               << SimpleParser::getString(arguments, "hw-address");

            std::string client_id;
            if (getOptionalString(arguments, "client-id", client_id)) {
                os << ", client-id: " << client_id;
                // It is not uncommon to provide a printable client ID
                try {
                    auto bin = ClientId::fromText(client_id)->getClientId();
                    if (str::isPrintable(bin)) {
                        os << " (" << BackendStore::vectorDump(bin) << ")";
                    }
                } catch (...) {
                    // Ignore any error
                }
            }

            addDuration(os, arguments);
            addContext(os, arguments);
        } else if ((name == "lease6-add") || name == "lease6-update") {
            if (name == "lease6-add") {
                os << " added a lease of";
            } else {
                os << " updated information on the lease of";
            }

            osa << SimpleParser::getString(arguments, "ip-address");
            if (isPrefix(arguments)) {
                os << " prefix: "
                   << SimpleParser::getString(arguments, "ip-address")
                   << "/"
                   << SimpleParser::getInteger(arguments, "prefix-len");
                osa << "/"
                    << SimpleParser::getInteger(arguments, "prefix-len");
            } else {
                os << " address: "
                    << SimpleParser::getString(arguments, "ip-address");
            }

            os  << " to a device with DUID: "
               << SimpleParser::getString(arguments, "duid");

            std::string hw_address;
            if (getOptionalString(arguments, "hw-address", hw_address)) {
                os << ", hardware address: " << hw_address;
            }

            addDuration(os, arguments);
            addContext(os, arguments);
        } else if ((name == "lease4-del") || (name == "lease6-del")) {
            std::string ip_address;
            if (getOptionalString(arguments, "ip-address", ip_address)) {
                osa << SimpleParser::getString(arguments, "ip-address");
                os << " deleted the lease for address: "
                   << SimpleParser::getString(arguments, "ip-address");
            } else {
                os << " deleted a lease for a device identified by: "
                   << SimpleParser::getString(arguments, "identifier-type")
                   << " of " << SimpleParser::getString(arguments, "identifier");
            }
        } else {
            // Don't care about any others
            return (0);
        }

        BackendStore::instance()->writeln(os.str(), osa.str());
    } catch (const std::exception& ex) {
        LOG_ERROR(legal_log_logger, LEGAL_LOG_COMMAND_WRITE_ERROR)
                  .arg(ex.what());
        return (1);
    }

    return (0);

}

} // end extern "C"
